(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
	typeof define === 'function' && define.amd ? define(['exports'], factory) :
	(global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global.JsonNode = {}));
})(this, (function (exports) { 'use strict';

	var TOKEN_NULL = 0x00;
	var TOKEN_SPACE = 0x01;
	var TOKEN_QUOTE = 0x04;
	var TOKEN_STRING = 0x05;
	var TOKEN_IDENT = 0x06;
	var TOKEN_NUMBER = 0x07;
	var TOKEN_SQUARE_OPEN = 0x41;
	var TOKEN_SQUARE_CLOSE = 0x42;
	var TOKEN_ROUND_OPEN = 0x43;
	var TOKEN_ROUND_CLOSE = 0x44;
	var TOKEN_GT = 0x45;
	var TOKEN_COLON = 0x46;
	var TOKEN_DOT = 0x47;
	var TOKEN_COMMA = 0x48;
	var TOKEN_HASH = 0x49;
	var TOKEN_EQUAL = 0x50;
	var TOKEN_ASTERISK = 0x51;
	var TOKEN_PLUS = 0x52;
	var TOKEN_TILDE = 0x53;
	var TOKEN_MINUS = 0x54;
	var TOKEN_PIPE = 0x55;
	var TOKEN_CARET_EQUAL = 0x61;
	var TOKEN_DOLLAR_EQUAL = 0x62;
	var TOKEN_ASTERISK_EQUAL = 0x63;
	var TOKEN_TILDE_EQUAL = 0x64;
	var TOKEN_PIPE_EQUAL = 0x65;
	var SINGLE_CHAR_OPERATORS = new Map([
	    ['[', TOKEN_SQUARE_OPEN],
	    [']', TOKEN_SQUARE_CLOSE],
	    ['(', TOKEN_ROUND_OPEN],
	    [')', TOKEN_ROUND_CLOSE],
	    ['>', TOKEN_GT],
	    [':', TOKEN_COLON],
	    ['.', TOKEN_DOT],
	    [',', TOKEN_COMMA],
	    ['#', TOKEN_HASH],
	    ['=', TOKEN_EQUAL],
	    ['*', TOKEN_ASTERISK],
	    ['+', TOKEN_PLUS],
	    ['~', TOKEN_TILDE],
	    ['-', TOKEN_MINUS],
	    ['|', TOKEN_PIPE],
	]);
	var DOUBLE_CHAR_OPERATORS = new Map([
	    ['^=', TOKEN_CARET_EQUAL],
	    ['$=', TOKEN_DOLLAR_EQUAL],
	    ['*=', TOKEN_ASTERISK_EQUAL],
	    ['~=', TOKEN_TILDE_EQUAL],
	    ['|=', TOKEN_PIPE_EQUAL],
	]);
	var SPACE_BEFORE_OPERATORS = [
	    TOKEN_GT,
	    TOKEN_COMMA,
	    TOKEN_PLUS,
	    TOKEN_MINUS,
	    TOKEN_TILDE,
	    TOKEN_ROUND_CLOSE,
	    TOKEN_SQUARE_CLOSE,
	];
	var SPACE_AFTER_OPERATORS = [
	    TOKEN_GT,
	    TOKEN_COMMA,
	    TOKEN_PLUS,
	    TOKEN_MINUS,
	    TOKEN_TILDE,
	    TOKEN_ROUND_OPEN,
	    TOKEN_SQUARE_OPEN,
	    TOKEN_CARET_EQUAL,
	    TOKEN_DOLLAR_EQUAL,
	    TOKEN_ASTERISK_EQUAL,
	    TOKEN_TILDE_EQUAL,
	    TOKEN_PIPE_EQUAL,
	];
	function isSpace(code) {
	    return code === 32 || code === 10 || code === 13 || code === 9 || code === 12;
	}
	function isDigit(code) {
	    return code >= 48 && code <= 57;
	}
	function isReplacer(code) {
	    return code === 65533;
	}
	function isHex(code) {
	    return isDigit(code) || (code >= 65 && code <= 70) || (code >= 97 && code <= 102);
	}
	function isUnicode(code) {
	    return code >= 160 && code < 65519;
	}
	function isIdentStart(code) {
	    return (code >= 65 && code <= 90) || code === 92 || code === 95 || (code >= 97 && code <= 122) || isUnicode(code);
	}
	function isIdentChar(code) {
	    return isDigit(code) || isIdentStart(code);
	}

	function tokenize(source) {
	    var length = source.length;
	    var tokens = [];
	    var cursor = 0;
	    while (cursor < length) {
	        var char = source.charAt(cursor);
	        if (char === '') {
	            cursor++;
	            continue;
	        }
	        if (char.trim() === '') {
	            var end = length - source.substring(cursor).trimStart().length;
	            tokens.push({
	                text: source.substring(cursor, end),
	                type: TOKEN_SPACE,
	                start: cursor,
	                end: end,
	            });
	            cursor = end;
	            continue;
	        }
	        if (char === '"' || char === "'") {
	            var end = source.indexOf(char, cursor + 1);
	            var hasEndQuote = end !== -1;
	            if (!hasEndQuote)
	                end = source.length;
	            tokens.push({
	                text: char,
	                type: TOKEN_QUOTE,
	                start: cursor,
	                end: cursor + 1,
	            });
	            var subIndex = cursor + 1;
	            while (subIndex < end) {
	                var subChar = source.charAt(subIndex);
	                var subEnd = subIndex + 1;
	                if (subChar === '\\' && subIndex + 1 < end)
	                    ++subEnd;
	                subIndex = subEnd;
	            }
	            tokens.push({
	                text: source.substring(cursor + 1, subIndex),
	                type: TOKEN_STRING,
	                start: cursor + 1,
	                end: subIndex,
	            });
	            cursor = end;
	            if (hasEndQuote) {
	                tokens.push({
	                    text: source.charAt(cursor),
	                    type: TOKEN_QUOTE,
	                    start: cursor,
	                    end: cursor + 1,
	                });
	                cursor++;
	            }
	            continue;
	        }
	        {
	            var isOperator = false;
	            if (cursor + 1 < length) {
	                var char2 = char + source.charAt(cursor + 1);
	                var tokenType = DOUBLE_CHAR_OPERATORS.get(char2);
	                if (tokenType != null) {
	                    var end = cursor + 2;
	                    tokens.push({
	                        text: char2,
	                        type: tokenType,
	                        start: cursor,
	                        end: end,
	                    });
	                    cursor = end;
	                    isOperator = true;
	                }
	            }
	            if (isOperator)
	                continue;
	            {
	                var tokenType = SINGLE_CHAR_OPERATORS.get(char);
	                if (tokenType != null) {
	                    var end = cursor + 1;
	                    tokens.push({
	                        text: char,
	                        type: tokenType,
	                        start: cursor,
	                        end: end,
	                    });
	                    cursor = end;
	                    isOperator = true;
	                }
	            }
	            if (isOperator)
	                continue;
	        }
	        {
	            var end = cursor;
	            if (isIdentStart(char.charCodeAt(0))) {
	                while (end < length && isIdentChar((char = source.charAt(end)).charCodeAt(0))) {
	                    if (char === '\\')
	                        ++end;
	                    ++end;
	                }
	                if (end > cursor) {
	                    var text = source.substring(cursor, end);
	                    tokens.push({
	                        text: text,
	                        type: TOKEN_IDENT,
	                        start: cursor,
	                        end: end,
	                    });
	                    cursor = end;
	                    continue;
	                }
	            }
	        }
	        {
	            var end = cursor;
	            if (isDigit(char.charCodeAt(0))) {
	                while (end < length && isDigit((char = source.charAt(end)).charCodeAt(0)))
	                    ++end;
	                if (end > cursor) {
	                    var text = source.substring(cursor, end);
	                    tokens.push({
	                        text: text,
	                        type: TOKEN_NUMBER,
	                        start: cursor,
	                        end: end,
	                    });
	                    cursor = end;
	                    continue;
	                }
	            }
	        }
	        {
	            var end = cursor + 1;
	            var prev = tokens[tokens.length - 1];
	            if (prev != null && prev.type === TOKEN_NULL) {
	                prev.text += char;
	                prev.end = end;
	            }
	            else {
	                tokens.push({
	                    text: char,
	                    type: TOKEN_NULL,
	                    start: cursor,
	                    end: end,
	                });
	            }
	            cursor = end;
	        }
	    }
	    return tokens.filter(function (token, i, arr) {
	        if (token.type === TOKEN_SPACE) {
	            var nextToken = arr[i + 1];
	            if (nextToken) {
	                if (SPACE_BEFORE_OPERATORS.includes(nextToken.type)) {
	                    nextToken.text = token.text + nextToken.text;
	                    nextToken.start = token.start;
	                    return false;
	                }
	            }
	            var prevToken = arr[i - 1];
	            if (prevToken) {
	                if (SPACE_AFTER_OPERATORS.includes(prevToken.type)) {
	                    prevToken.text = prevToken.text + token.text;
	                    prevToken.end = token.end;
	                    return false;
	                }
	            }
	        }
	        return true;
	    });
	}

	/******************************************************************************
	Copyright (c) Microsoft Corporation.

	Permission to use, copy, modify, and/or distribute this software for any
	purpose with or without fee is hereby granted.

	THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
	REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
	AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
	INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
	LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
	OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
	PERFORMANCE OF THIS SOFTWARE.
	***************************************************************************** */

	function __generator(thisArg, body) {
	    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
	    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
	    function verb(n) { return function (v) { return step([n, v]); }; }
	    function step(op) {
	        if (f) throw new TypeError("Generator is already executing.");
	        while (_) try {
	            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
	            if (y = 0, t) op = [op[0] & 2, t.value];
	            switch (op[0]) {
	                case 0: case 1: t = op; break;
	                case 4: _.label++; return { value: op[1], done: false };
	                case 5: _.label++; y = op[1]; op = [0]; continue;
	                case 7: op = _.ops.pop(); _.trys.pop(); continue;
	                default:
	                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
	                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
	                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
	                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
	                    if (t[2]) _.ops.pop();
	                    _.trys.pop(); continue;
	            }
	            op = body.call(thisArg, _);
	        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
	        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
	    }
	}

	function __values(o) {
	    var s = typeof Symbol === "function" && Symbol.iterator, m = s && o[s], i = 0;
	    if (m) return m.call(o);
	    if (o && typeof o.length === "number") return {
	        next: function () {
	            if (o && i >= o.length) o = void 0;
	            return { value: o && o[i++], done: !o };
	        }
	    };
	    throw new TypeError(s ? "Object is not iterable." : "Symbol.iterator is not defined.");
	}

	function __spreadArray(to, from, pack) {
	    if (pack || arguments.length === 2) for (var i = 0, l = from.length, ar; i < l; i++) {
	        if (ar || !(i in from)) {
	            if (!ar) ar = Array.prototype.slice.call(from, 0, i);
	            ar[i] = from[i];
	        }
	    }
	    return to.concat(ar || Array.prototype.slice.call(from));
	}

	function _isNodeInternal(value) {
	    return typeof value !== 'string';
	}

	function isSelectableNode(value) {
	    return _isNodeInternal(value) ? value.tagName !== '' && !value.tagName.startsWith('?') && !value.tagName.startsWith('!') : false;
	}

	function childNodes(node) {
	    return node.children.filter(isSelectableNode);
	}

	function getAttribute(node, name, defaults) {
	    var _a, _b;
	    return (_b = (_a = node.attributes.find(function (attr) { return attr.name === name; })) === null || _a === void 0 ? void 0 : _a.value) !== null && _b !== void 0 ? _b : defaults;
	}

	function hasAttribute(node, name) {
	    return node.attributes.some(function (attr) { return attr.name === name; });
	}

	function indexOf(node) {
	    return node.parent == null ? 0 : childNodes(node.parent).indexOf(node);
	}

	function siblings(node) {
	    if (node.parent == null)
	        return [node];
	    return childNodes(node.parent);
	}

	function indexOfType(node) {
	    return node.parent == null
	        ? 0
	        : siblings(node)
	            .filter(function (sibling) { return sibling.tagName === node.tagName; })
	            .indexOf(node);
	}

	function lastIndexOf(node) {
	    if (node.parent == null)
	        return 0;
	    var nodes = childNodes(node.parent);
	    return nodes.length - nodes.indexOf(node) - 1;
	}

	function lastIndexOfType(node) {
	    if (node.parent == null)
	        return 0;
	    var nodes = siblings(node).filter(function (sibling) { return sibling.tagName === node.tagName; });
	    return nodes.length - nodes.indexOf(node) - 1;
	}

	function parents(node, excludeSelf) {
	    if (excludeSelf === void 0) { excludeSelf = false; }
	    var list = [];
	    var current = node;
	    if (excludeSelf)
	        current = current.parent;
	    while (current) {
	        list.push(current);
	        current = current.parent;
	    }
	    return list;
	}

	function previousSibling(node) {
	    if (node.parent == null)
	        return null;
	    var siblingNodes = childNodes(node.parent);
	    var nodeIndex = siblingNodes.indexOf(node);
	    return nodeIndex === 0 ? null : siblingNodes[nodeIndex - 1];
	}

	function isComment(value) {
	    return typeof value === 'string' && value.startsWith('<!--') && value.endsWith('-->');
	}

	function isNode(value) {
	    return (typeof value === 'object' &&
	        Object.prototype.hasOwnProperty.call(value, 'tagName') &&
	        Object.prototype.hasOwnProperty.call(value, 'attributes') &&
	        Object.prototype.hasOwnProperty.call(value, 'children'));
	}

	function textContent(node) {
	    return node.children.map(function (n) { return (isNode(n) ? textContent(n) : isComment(n) ? '' : n); }).join('');
	}

	function createSyntaxError(token) {
	    if (token == null)
	        return new SyntaxError("Unexpected end of input");
	    return new SyntaxError("Unexpected token \"".concat(token.text, "\" at position ").concat(token.start));
	}

	function unescape(text) {
	    var length = text.length;
	    var result = '';
	    var cursor = 0;
	    while (cursor < length) {
	        var char = text.charAt(cursor++);
	        if (char === '\\') {
	            var hex = '';
	            while (isHex((char = text.charAt(cursor)).charCodeAt(0))) {
	                hex += char;
	                ++cursor;
	            }
	            if (hex === '') {
	                result += char;
	                ++cursor;
	            }
	            else {
	                result += String.fromCharCode(parseInt(hex, 16));
	                if (isSpace(char.charCodeAt(0)))
	                    ++cursor;
	            }
	        }
	        else if (isReplacer(char.charCodeAt(0))) {
	            result += '\0';
	        }
	        else {
	            result += char;
	        }
	    }
	    return result;
	}

	function parseTokens(tokens) {
	    var groupedCheckers = [];
	    var chainedCheckers = [];
	    var checkers = [];
	    var length = tokens.length + 1;
	    var cursor = 0;
	    var skip = function () {
	        var _a;
	        var types = [];
	        for (var _i = 0; _i < arguments.length; _i++) {
	            types[_i] = arguments[_i];
	        }
	        if (types.includes((_a = tokens[cursor]) === null || _a === void 0 ? void 0 : _a.type)) {
	            ++cursor;
	            return true;
	        }
	        return false;
	    };
	    var is = function () {
	        var types = [];
	        for (var _i = 0; _i < arguments.length; _i++) {
	            types[_i] = arguments[_i];
	        }
	        var next = tokens[cursor];
	        return next && types.includes(next.type) ? next : null;
	    };
	    var not = function () {
	        var types = [];
	        for (var _i = 0; _i < arguments.length; _i++) {
	            types[_i] = arguments[_i];
	        }
	        var next = tokens[cursor];
	        return next && !types.includes(next.type) ? next : null;
	    };
	    var nextName = function () {
	        var extraTokens = [];
	        for (var _i = 0; _i < arguments.length; _i++) {
	            extraTokens[_i] = arguments[_i];
	        }
	        var start = null;
	        var end = null;
	        var text = '';
	        var next = is.apply(void 0, __spreadArray([TOKEN_IDENT, TOKEN_MINUS], extraTokens, false));
	        while (next) {
	            if (start == null)
	                start = next.start;
	            end = next.end;
	            text += next.text;
	            ++cursor;
	            next = is.apply(void 0, __spreadArray([TOKEN_IDENT, TOKEN_MINUS], extraTokens, false));
	        }
	        return text.length === 0 || start == null || end == null
	            ? null
	            : {
	                type: TOKEN_IDENT,
	                text: text,
	                start: start,
	                end: end,
	            };
	    };
	    var _loop_1 = function () {
	        var token = tokens[cursor++];
	        if (token == null || token.type === TOKEN_COMMA) {
	            if (checkers.length > 0) {
	                var grouped = groupCheckers(checkers);
	                if (grouped) {
	                    chainedCheckers.unshift([TOKEN_SPACE, grouped]);
	                    checkers = [];
	                }
	            }
	            var chained = chainCheckers(chainedCheckers);
	            if (chained) {
	                groupedCheckers.push(chained);
	                chainedCheckers = [];
	                return "continue";
	            }
	        }
	        else if (token.type === TOKEN_SPACE || token.type === TOKEN_GT || token.type === TOKEN_PLUS || token.type === TOKEN_TILDE) {
	            var grouped = groupCheckers(checkers);
	            if (grouped) {
	                chainedCheckers.unshift([token.type, grouped]);
	                checkers = [];
	                return "continue";
	            }
	            if (token.type === TOKEN_SPACE) {
	                return "continue";
	            }
	        }
	        else if (token.type === TOKEN_COLON) {
	            var extraColon = is(TOKEN_COLON) ? token : null;
	            if (extraColon)
	                cursor++;
	            var nameToken = nextName();
	            if (nameToken) {
	                var pseudoName = nameToken.text;
	                if (pseudoName === 'root') {
	                    checkers.push(function (node) { return node.parent == null || node.parent.tagName === ''; });
	                    return "continue";
	                }
	                else if (pseudoName === 'scope') {
	                    checkers.push(function (node, scope) { return node === scope; });
	                    return "continue";
	                }
	                else if (pseudoName === 'empty') {
	                    checkers.push(function (node) { return childNodes(node).length === 0 && textContent(node).length === 0; });
	                    return "continue";
	                }
	                else if (pseudoName === 'disabled') {
	                    checkers.push(function (node) { var _a; return ['disabled', 'true'].includes((_a = getAttribute(node, 'disabled')) !== null && _a !== void 0 ? _a : ''); });
	                    return "continue";
	                }
	                else if (pseudoName === 'enabled') {
	                    checkers.push(function (node) { var _a; return !['disabled', 'true'].includes((_a = getAttribute(node, 'disabled')) !== null && _a !== void 0 ? _a : ''); });
	                    return "continue";
	                }
	                else if (pseudoName === 'first-child') {
	                    checkers.push(function (node) { return indexOf(node) === 0; });
	                    return "continue";
	                }
	                else if (pseudoName === 'last-child') {
	                    checkers.push(function (node) { return lastIndexOf(node) === 0; });
	                    return "continue";
	                }
	                else if (pseudoName === 'only-child') {
	                    checkers.push(function (node) { return node.parent == null || childNodes(node.parent).length === 1; });
	                    return "continue";
	                }
	                else if (pseudoName === 'first-of-type') {
	                    checkers.push(function (node) { return indexOfType(node) === 0; });
	                    return "continue";
	                }
	                else if (pseudoName === 'last-of-type') {
	                    checkers.push(function (node) { return lastIndexOfType(node) === 0; });
	                    return "continue";
	                }
	                else if (pseudoName === 'only-of-type') {
	                    checkers.push(function (node) { return indexOfType(node) === 0 && lastIndexOfType(node) === 0; });
	                    return "continue";
	                }
	                else if (['first-letter', 'first-line', 'before', 'after'].includes(pseudoName)) {
	                    checkers.push(function () { return false; });
	                    return "continue";
	                }
	                var pseudoParam = [];
	                skip(TOKEN_ROUND_OPEN);
	                var paramToken = void 0;
	                while ((paramToken = not(TOKEN_ROUND_CLOSE))) {
	                    ++cursor;
	                    pseudoParam.push(paramToken);
	                }
	                skip(TOKEN_ROUND_CLOSE);
	                if (pseudoName === 'nth-child') {
	                    var expr_1 = parseExpression(pseudoParam);
	                    if (expr_1) {
	                        checkers.push(function (node) { return expr_1(indexOf(node) + 1); });
	                        return "continue";
	                    }
	                }
	                else if (pseudoName === 'nth-last-child') {
	                    var expr_2 = parseExpression(pseudoParam);
	                    if (expr_2) {
	                        checkers.push(function (node) { return expr_2(lastIndexOf(node) + 1); });
	                        return "continue";
	                    }
	                }
	                else if (pseudoName === 'nth-of-type') {
	                    var expr_3 = parseExpression(pseudoParam);
	                    if (expr_3) {
	                        checkers.push(function (node) { return expr_3(indexOfType(node) + 1); });
	                        return "continue";
	                    }
	                }
	                else if (pseudoName === 'nth-last-of-type') {
	                    var expr_4 = parseExpression(pseudoParam);
	                    if (expr_4) {
	                        checkers.push(function (node) { return expr_4(lastIndexOfType(node) + 1); });
	                        return "continue";
	                    }
	                }
	                else if (pseudoName === 'not') {
	                    var notChecker_1 = parseTokens(pseudoParam);
	                    if (notChecker_1) {
	                        checkers.push(function (node, scope) { return !notChecker_1(node, scope); });
	                        return "continue";
	                    }
	                }
	                else if (['slotted'].includes(pseudoName)) {
	                    checkers.push(function () { return false; });
	                    return "continue";
	                }
	                else {
	                    throw createSyntaxError(extraColon !== null && extraColon !== void 0 ? extraColon : nameToken);
	                }
	            }
	        }
	        else if (token.type === TOKEN_SQUARE_OPEN) {
	            var nameToken = nextName(TOKEN_COLON);
	            if (nameToken) {
	                var attrName_1 = nameToken.text;
	                var attrValue = null;
	                var cmpToken = not(TOKEN_SQUARE_CLOSE);
	                if (cmpToken) {
	                    ++cursor;
	                    skip(TOKEN_QUOTE);
	                    var valueToken = is(TOKEN_STRING, TOKEN_IDENT);
	                    if (valueToken) {
	                        ++cursor;
	                        attrValue = valueToken.text;
	                    }
	                    else
	                        attrValue = '';
	                    skip(TOKEN_QUOTE);
	                }
	                if (attrValue == null) {
	                    checkers.push(function (node) { return hasAttribute(node, attrName_1); });
	                }
	                else {
	                    var attrValueUnescape_1 = unescape(attrValue);
	                    if (cmpToken != null) {
	                        if (cmpToken.type === TOKEN_EQUAL) {
	                            checkers.push(function (node) { return getAttribute(node, attrName_1) === attrValueUnescape_1; });
	                        }
	                        else if (cmpToken.type === TOKEN_CARET_EQUAL) {
	                            checkers.push(function (node) { var _a; return attrValueUnescape_1 !== '' && !!((_a = getAttribute(node, attrName_1)) === null || _a === void 0 ? void 0 : _a.startsWith(attrValueUnescape_1)); });
	                        }
	                        else if (cmpToken.type === TOKEN_DOLLAR_EQUAL) {
	                            checkers.push(function (node) { var _a; return attrValueUnescape_1 !== '' && !!((_a = getAttribute(node, attrName_1)) === null || _a === void 0 ? void 0 : _a.endsWith(attrValueUnescape_1)); });
	                        }
	                        else if (cmpToken.type === TOKEN_ASTERISK_EQUAL) {
	                            attrValueUnescape_1 = attrValueUnescape_1.trim();
	                            checkers.push(function (node) { var _a; return attrValueUnescape_1 !== '' && !!((_a = getAttribute(node, attrName_1)) === null || _a === void 0 ? void 0 : _a.includes(attrValueUnescape_1)); });
	                        }
	                        else if (cmpToken.type === TOKEN_TILDE_EQUAL) {
	                            checkers.push(function (node) {
	                                if (attrValueUnescape_1 === '')
	                                    return false;
	                                var list = getAttribute(node, attrName_1, '').split(' ');
	                                return list.length > 0 && list.includes(attrValueUnescape_1);
	                            });
	                        }
	                        else if (cmpToken.type === TOKEN_PIPE_EQUAL) {
	                            checkers.push(function (node) {
	                                if (attrValueUnescape_1 === '')
	                                    return false;
	                                var list = getAttribute(node, attrName_1, '').split('-');
	                                return list.length > 0 && list.includes(attrValueUnescape_1);
	                            });
	                        }
	                        else
	                            throw createSyntaxError(cmpToken);
	                    }
	                }
	                if (skip(TOKEN_SQUARE_CLOSE)) {
	                    return "continue";
	                }
	                if (tokens[cursor + 1] == null)
	                    return "continue";
	            }
	        }
	        else if (token.type === TOKEN_IDENT) {
	            --cursor;
	            var nameToken = nextName();
	            if (nameToken) {
	                var nodeName_1 = unescape(nameToken.text);
	                checkers.push(function (node) { return node.tagName === nodeName_1; });
	                return "continue";
	            }
	        }
	        else if (token.type === TOKEN_HASH) {
	            var nameToken = nextName(TOKEN_NUMBER);
	            if (nameToken) {
	                var nodeName_2 = unescape(nameToken.text);
	                checkers.push(function (node) { return getAttribute(node, 'id') === nodeName_2; });
	                return "continue";
	            }
	        }
	        else if (token.type === TOKEN_DOT) {
	            var nameToken = nextName();
	            if (nameToken) {
	                var nodeName_3 = unescape(nameToken.text);
	                checkers.push(function (node) { return getAttribute(node, 'class', '').split(' ').includes(nodeName_3); });
	                return "continue";
	            }
	        }
	        else if (token.type === TOKEN_ASTERISK) {
	            if (checkers.length === 0) {
	                checkers.unshift(function (node) { return node.tagName !== ''; });
	            }
	            return "continue";
	        }
	        throw createSyntaxError(token);
	    };
	    while (cursor < length) {
	        _loop_1();
	    }
	    if (groupedCheckers.length === 0) {
	        throw new SyntaxError("The provided selector is empty.");
	    }
	    return function (node, scope) { return groupedCheckers.some(function (check) { return check(node, scope); }); };
	}
	function groupCheckers(checkers) {
	    if (checkers.length) {
	        return function (node, scope) { return checkers.every(function (check) { return check(node, scope); }); };
	    }
	    return null;
	}
	function chainCheckers(checkers) {
	    if (checkers.length === 0)
	        return null;
	    if (checkers.length === 1)
	        return checkers[0][1];
	    return function (node, scope) {
	        if (!checkers[0][1](node, scope))
	            return false;
	        var _loop_2 = function (index, parent) {
	            var _a = checkers[index], type = _a[0], check = _a[1];
	            if (parent.parent) {
	                if (type === TOKEN_SPACE) {
	                    parent = parents(parent, true).find(function (node2) { return check(node2, scope); });
	                }
	                else if (type === TOKEN_GT) {
	                    parent = check(parent.parent, scope) ? parent.parent : undefined;
	                }
	                else if (type === TOKEN_PLUS) {
	                    var prev = previousSibling(parent);
	                    parent = prev && check(prev, scope) ? prev : undefined;
	                }
	                else if (type === TOKEN_TILDE) {
	                    var nodeIndex = indexOf(parent);
	                    if (nodeIndex === 0)
	                        parent = undefined;
	                    else {
	                        var precursors = childNodes(parent.parent).slice(0, nodeIndex);
	                        parent = precursors.find(function (node2) { return check(node2, scope); });
	                    }
	                }
	            }
	            else
	                parent = undefined;
	            if (parent != null && index === checkers.length - 1)
	                return { value: true };
	            out_parent_1 = parent;
	        };
	        var out_parent_1;
	        for (var index = 1, parent = node; parent != null; index++) {
	            var state_1 = _loop_2(index, parent);
	            parent = out_parent_1;
	            if (typeof state_1 === "object")
	                return state_1.value;
	        }
	        return false;
	    };
	}
	function parseExpression(expr) {
	    if (expr.length === 0)
	        return null;
	    if (expr.length <= 5) {
	        var pos = 0;
	        var multiplier_1 = 1;
	        var offset_1 = 0;
	        var firstSign = expr[pos] && (expr[pos].type === TOKEN_PLUS || expr[pos].type === TOKEN_MINUS) ? expr[pos].type : null;
	        if (firstSign)
	            pos++;
	        var firstNumber_1 = undefined;
	        if (expr[pos] && expr[pos].type === TOKEN_NUMBER) {
	            firstNumber_1 = parseInt(expr[pos].text, 10);
	            pos++;
	        }
	        var hasVariable = expr[pos] && expr[pos].type === TOKEN_IDENT;
	        if (hasVariable) {
	            if (expr[pos].text !== 'n')
	                throw createSyntaxError(expr[pos]);
	            pos++;
	        }
	        var secondSign = expr[pos] && (expr[pos].type === TOKEN_PLUS || expr[pos].type === TOKEN_MINUS) ? expr[pos].type : null;
	        if (secondSign)
	            pos++;
	        var secondNumber = undefined;
	        if (secondSign != null) {
	            if (expr[pos] && expr[pos].type === TOKEN_NUMBER) {
	                secondNumber = parseInt(expr[pos].text, 10);
	                pos++;
	            }
	            else {
	                throw createSyntaxError(expr[pos]);
	            }
	        }
	        if (pos !== expr.length)
	            throw createSyntaxError(expr[pos]);
	        if (hasVariable) {
	            multiplier_1 = firstNumber_1 !== null && firstNumber_1 !== void 0 ? firstNumber_1 : 1;
	        }
	        if (secondSign != null) {
	            if (secondSign === TOKEN_PLUS) {
	                offset_1 = secondNumber !== null && secondNumber !== void 0 ? secondNumber : 0;
	            }
	            else if (secondSign === TOKEN_MINUS) {
	                offset_1 = -(secondNumber !== null && secondNumber !== void 0 ? secondNumber : 0);
	            }
	        }
	        if (hasVariable) {
	            if (firstSign === TOKEN_MINUS) {
	                return function (index) {
	                    index -= offset_1;
	                    return index <= 0 && index % multiplier_1 === 0;
	                };
	            }
	            else {
	                return function (index) {
	                    index -= offset_1;
	                    return index >= 0 && index % multiplier_1 === 0;
	                };
	            }
	        }
	        else {
	            return function (index) { return index === firstNumber_1; };
	        }
	    }
	    throw createSyntaxError(expr[5]);
	}

	var cache = new Map();
	var Selector = {
	    unescape: unescape,
	    parse: function (selector) {
	        if (cache.has(selector))
	            return cache.get(selector);
	        var tokens = tokenize(selector);
	        var check = parseTokens(tokens);
	        cache.set(selector, check);
	        return check;
	    },
	};

	function createAttrs(attrsMap) {
	    return attrsMap ? Object.keys(attrsMap).map(function (attr) { return ({ name: attr, value: attrsMap[attr] }); }) : [];
	}

	function createAttrsMap(attrs) {
	    var attrsMap = {};
	    if (attrs)
	        attrs.forEach(function (attr) { return void (attrsMap[attr.name] = attr.value); });
	    return attrsMap;
	}

	function createNode(tagName, attrsMap) {
	    if (tagName === void 0) { tagName = ''; }
	    return {
	        tagName: tagName,
	        attributes: createAttrs(attrsMap),
	        parent: null,
	        children: [],
	    };
	}

	function toJson(node) {
	    var result = { name: node.tagName };
	    if (node.attributes.length)
	        result.attr = createAttrsMap(node.attributes);
	    if (node.children.length)
	        result.child = node.children.map(function (child) { return (_isNodeInternal(child) ? toJson(child) : child); });
	    return result;
	}

	function removeChild(node, child) {
	    if (node !== child) {
	        var index = node.children.indexOf(child);
	        if (index >= 0)
	            node.children.splice(index, 1);
	        child.parent = null;
	    }
	    return node;
	}

	function appendChild(node, child) {
	    if (_isNodeInternal(child)) {
	        if (node !== child) {
	            if (child.parent)
	                removeChild(child.parent, child);
	            child.parent = node;
	            node.children.push(child);
	        }
	    }
	    else {
	        node.children.push(child);
	    }
	}

	function fromJson(json) {
	    var _a;
	    var node = createNode((_a = json.name) !== null && _a !== void 0 ? _a : '', json.attr);
	    if (json.child) {
	        json.child.forEach(function (child) {
	            appendChild(node, typeof child === 'string' ? child : fromJson(child));
	        });
	    }
	    return node;
	}

	function cloneAttr(attr) {
	    return {
	        name: attr.name,
	        value: attr.value,
	    };
	}

	function cloneNode(node, deep) {
	    if (deep === void 0) { deep = false; }
	    var clone = createNode(node.tagName, createAttrsMap(node.attributes));
	    if (deep) {
	        node.children.forEach(function (child) {
	            appendChild(clone, _isNodeInternal(child) ? cloneNode(child, true) : child);
	        });
	    }
	    return clone;
	}

	function setAttribute(node, name, value) {
	    var old = node.attributes.find(function (attr) { return attr.name === name; });
	    if (old)
	        old.value = value;
	    else
	        node.attributes.push({ name: name, value: value });
	}

	function copyAttr(from, to) {
	    from.attributes.forEach(function (a) { return setAttribute(to, a.name, a.value); });
	    return to;
	}

	function copyChild(from, to) {
	    from.children.forEach(function (n) { return appendChild(to, _isNodeInternal(n) ? cloneNode(n, true) : n); });
	    return to;
	}

	function copyNode(from) {
	    return copyAttr(from, createNode(from.tagName));
	}

	function removeAttribute(node, name) {
	    var index = node.attributes.findIndex(function (attr) { return attr.name === name; });
	    if (index >= 0)
	        node.attributes.splice(index, 1);
	}

	function coverAttr(node, form, to) {
	    if (!hasAttribute(node, to) && hasAttribute(node, form)) {
	        setAttribute(node, to, getAttribute(node, form, ''));
	        removeAttribute(node, form);
	    }
	}

	function descendants(node) {
	    var list = [];
	    for (var _i = 0, _a = node.children; _i < _a.length; _i++) {
	        var child = _a[_i];
	        if (isSelectableNode(child))
	            list.push.apply(list, __spreadArray([child], descendants(child), false));
	    }
	    return list;
	}

	function descendantsIt(node) {
	    var _i, _a, child;
	    return __generator(this, function (_b) {
	        switch (_b.label) {
	            case 0:
	                _i = 0, _a = node.children;
	                _b.label = 1;
	            case 1:
	                if (!(_i < _a.length)) return [3 /*break*/, 5];
	                child = _a[_i];
	                if (!isSelectableNode(child)) return [3 /*break*/, 4];
	                return [4 /*yield*/, child];
	            case 2:
	                _b.sent();
	                return [5 /*yield**/, __values(descendants(child))];
	            case 3:
	                _b.sent();
	                _b.label = 4;
	            case 4:
	                _i++;
	                return [3 /*break*/, 1];
	            case 5: return [2 /*return*/];
	        }
	    });
	}

	function find(node, selector) {
	    return descendants(node).filter(function (node2) { return Selector.parse(selector)(node2, node); });
	}

	function findChild(node, selector) {
	    return childNodes(node).find(function (node2) { return Selector.parse(selector)(node2, node); });
	}

	function findChildren(node, selector) {
	    return childNodes(node).filter(function (node2) { return Selector.parse(selector)(node2, node); });
	}

	function findFirst(node, selector) {
	    return descendants(node).find(function (node2) { return Selector.parse(selector)(node2, node); });
	}

	function findParent(node, selector) {
	    return parents(node).find(function (node2) { return Selector.parse(selector)(node2, node); });
	}

	function findParents(node, selector) {
	    return parents(node).filter(function (node2) { return Selector.parse(selector)(node2, node); });
	}

	function firstChild(node) {
	    return node.children.find(isSelectableNode);
	}

	function insertAfter(node, child, ref) {
	    if (node !== child && child !== ref && ref.parent === node) {
	        if (child.parent)
	            removeChild(child.parent, child);
	        child.parent = node;
	        node.children.splice(node.children.indexOf(ref) + 1, 0, child);
	    }
	}

	function insertBefore(node, child, ref) {
	    if (node !== child && child !== ref && ref.parent === node) {
	        if (child.parent)
	            removeChild(child.parent, child);
	        child.parent = node;
	        node.children.splice(node.children.indexOf(ref), 0, child);
	    }
	}

	function nextSibling(node) {
	    if (node.parent == null)
	        return null;
	    var siblingNodes = childNodes(node.parent);
	    var nodeIndex = siblingNodes.indexOf(node);
	    return nodeIndex === siblingNodes.length - 1 ? null : siblingNodes[nodeIndex + 1];
	}

	function removeAttr(node, except) {
	    node.attributes
	        .map(function (n) { return n.name; })
	        .filter(function (a) { return except == null || !except.includes(a); })
	        .forEach(function (a) { return removeAttribute(node, a); });
	}

	function children(node) {
	    return __spreadArray([], node.children, true);
	}

	exports.Selector = Selector;
	exports.appendChild = appendChild;
	exports.childNodes = childNodes;
	exports.children = children;
	exports.cloneAttr = cloneAttr;
	exports.cloneNode = cloneNode;
	exports.copyAttr = copyAttr;
	exports.copyChild = copyChild;
	exports.copyNode = copyNode;
	exports.coverAttr = coverAttr;
	exports.createAttrs = createAttrs;
	exports.createAttrsMap = createAttrsMap;
	exports.createNode = createNode;
	exports.descendants = descendants;
	exports.descendantsIt = descendantsIt;
	exports.find = find;
	exports.findChild = findChild;
	exports.findChildren = findChildren;
	exports.findFirst = findFirst;
	exports.findParent = findParent;
	exports.findParents = findParents;
	exports.firstChild = firstChild;
	exports.fromJson = fromJson;
	exports.getAttribute = getAttribute;
	exports.hasAttribute = hasAttribute;
	exports.indexOf = indexOf;
	exports.indexOfType = indexOfType;
	exports.insertAfter = insertAfter;
	exports.insertBefore = insertBefore;
	exports.isComment = isComment;
	exports.isNode = isNode;
	exports.isSelectableNode = isSelectableNode;
	exports.lastIndexOf = lastIndexOf;
	exports.lastIndexOfType = lastIndexOfType;
	exports.nextSibling = nextSibling;
	exports.parents = parents;
	exports.previousSibling = previousSibling;
	exports.removeAttr = removeAttr;
	exports.removeAttribute = removeAttribute;
	exports.removeChild = removeChild;
	exports.setAttribute = setAttribute;
	exports.siblings = siblings;
	exports.textContent = textContent;
	exports.toJson = toJson;

	Object.defineProperty(exports, '__esModule', { value: true });

}));
//# sourceMappingURL=index.js.map
